package Networking;

import java.io.*;
import java.net.*;
import java.util.*;

public class ChatServer implements Runnable {

    public final static int DEFAULT_PORT = 2000;
    public final static int DEFAULT_MAX_BACKLOG = 5;
    public final static int DEFAULT_MAX_CONNECTIONS = 20;
    public final static String DEFAULT_HOST_FILE = "hosts.txt";
    public final static String MAGIC = "Yippy Skippy";
    private String magic = MAGIC;
    private int port = DEFAULT_PORT;
    private int backlog = DEFAULT_MAX_BACKLOG;
    private int numConnections = 0;
    private int maxConnections = DEFAULT_MAX_CONNECTIONS;
    private String hostfile = DEFAULT_HOST_FILE;
    private Vector connections = null;
    private Hashtable hostInfo = new Hashtable(15);

    public static void main(String[] args) {
        ChatServer cs = new ChatServer();
        cs.go();
    }

    /******************************************************************
    This method reads the command line properties that can be set
    by the user and then creates a thread to run the ChatServer.
     ******************************************************************/
    public void go() {

        // get the system properties specified by the user on the command line
        String portString = System.getProperty("port");
        if (portString != null) {
            port = Integer.parseInt(portString);
        }
        String backlogString = System.getProperty("backlog");
        if (backlogString != null) {
            backlog = Integer.parseInt(backlogString);
        }
        String hostFileString = System.getProperty("hostfile");
        if (hostFileString != null) {
            hostfile = hostFileString;
        }
        String magicString = System.getProperty("magic");
        if (magicString != null) {
            magic = magicString;
        }
        String connections = System.getProperty("connections");
        if (connections != null) {
            maxConnections = Integer.parseInt(connections);
        }
        this.connections = new Vector(maxConnections);

        System.out.println("Server settings:\n      Port=" + port + "\n Max Backlog=" +
                backlog + "\n      Max Connections=" + maxConnections +
                "\n        Host File=" + hostfile);

        // read the host file
        createHostList();

        this.port = port;

        // create a new thread and have it run the ChatServer
        Thread t = new Thread(this);
        t.start();
    }

    /******************************************************************
    This method reads the hosts file which contains a list of
    host names and student names.
    
    The information read from the file is stored in the hashtable
    hostInfo using the host name as the key and the student name as
    the value.
     *******************************************************************/
    private void createHostList() {

        File hostFile = new File(hostfile);
        try {
            System.out.println("Attempting to read hostfile hosts.txt: ");

            // create the input streams needed to read the hosts file
            FileInputStream fis = new FileInputStream(hostFile);
            InputStreamReader isr = new InputStreamReader(fis);
            BufferedReader br = new BufferedReader(isr);
            String aLine = null;

            // read a line from the file
            while ((aLine = br.readLine()) != null) {
                // search for the space that separates the host name from
                // the student name
                int spaceIndex = aLine.indexOf(' ');
                if (spaceIndex == -1) {
                    System.out.println("Invalid line in host file: " + aLine);
                    continue;
                }

                // get the host name
                String host = aLine.substring(0, spaceIndex);

                // get the student name
                String student = aLine.substring(spaceIndex + 1);
                System.out.println("Read: " + student + "@" + host);

                // store the student name in the hashtable using the host
                // name as the key
                hostInfo.put(host, student);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /******************************************************************
    This method creates a serverSocket and then listens for client
    connections.  After a client connects handleConnection gets called
    which creates a new thread to process the client's request while
    the server waits for other client connections.
     *******************************************************************/
    public void run() {


        ServerSocket serverSocket = null;
        Socket communicationSocket;

        try {
            System.out.println("Attempting to start server...");

            // create the ServerSocket
            serverSocket = new ServerSocket(port, backlog);
        } catch (IOException e) {
            System.out.println("Error starting server: Could not open port " + port);
            e.printStackTrace();
            System.exit(-1);
        }
        System.out.println("Started server on port " + port);

        // Run the listen/accept loop forever
        while (true) {
            try {
                // Wait here and listen for a connection
                communicationSocket = serverSocket.accept();

                // call handleConnection to process the client's connection
                handleConnection(communicationSocket);
            } catch (Exception e) {
                System.out.println("Unable to spawn child socket.");
                e.printStackTrace();
            }
        }
    }

    /******************************************************************
    This method creates a new thread to handle the client's Connection.
    It also adds the new connection to the list of connections to
    the chat server.
     ******************************************************************/
    public void handleConnection(Socket connection) {

        synchronized (this) {
            // If the maximum number of connections has been reached
            // make the client wait until a connection becomes available
            while (numConnections == maxConnections) {
                try {
                    wait();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }

        // If the maximum number of connections has not be reached or a client
        // has been notified that a connection is now available, increment
        // the number of connections
        numConnections++;

        //  create a new connection to handle client
        Connection con = new Connection(connection);

        // create a new thread to run the client's connection
        Thread t = new Thread(con);
        t.start();

        // add the new connection to the list of current connections
        // to the chat server
        connections.addElement(con);
    }

    /******************************************************************
    This method removes a connection from the connection list whenever
    a client closes a connection to the chat server.  It then notifies
    any waiting client's that a connection is available.
     ******************************************************************/
    public synchronized void connectionClosed(Connection connection) {

        // remove the connection from the list of connections
        connections.removeElement(connection);

        // decrement the current number of connections
        numConnections--;

        // notify any waiting client that a connection is now available
        notify();
    }

    /*******************************************************************
    This method sends a message to all clients connected to the
    chat server.
     *******************************************************************/
    public void sendToAllClients(String message) {

        // get a listing of all connections
        Enumeration cons = connections.elements();

        // loop through list of connections
        while (cons.hasMoreElements()) {
            // get next connection
            Connection c = (Connection) cons.nextElement();

            // send message
            c.sendMessage(message);
        }
    }

    /*******************************************************************
    This class represents a connection to a client and handles reading
    messages from the client.
     *******************************************************************/
    class Connection implements Runnable {

        private Socket communicationSocket = null;
        private OutputStreamWriter out = null;
        private BufferedReader in = null;

        /*******************************************************************
        This method handles reading a client's message and sending
        it out to all clients in the chat room.
         *******************************************************************/
        public void run() {
            OutputStream socketOutput = null;
            InputStream socketInput = null;

            try {
                // get the input and output streams from the socket
                socketOutput = communicationSocket.getOutputStream();
                out = new OutputStreamWriter(socketOutput);
                socketInput = communicationSocket.getInputStream();
                in = new BufferedReader(new InputStreamReader(socketInput));

                // get the client's hostname
                InetAddress address = communicationSocket.getInetAddress();
                String hostname = address.getHostName();

                // send welcome message to client
                String welcome =
                        "Connection made from host: " + hostname + "\nEverybody say hello";
                String student = (String) (hostInfo.get(hostname));
                if (student != null) {
                    welcome += " to " + student;
                }
                welcome += "!\n";

                // notify all client's that a new client has joined the session
                ChatServer.this.sendToAllClients(welcome);
                System.out.println("Connection made " + student + "@" + hostname);

                // tell client the "magic" phrase
                sendMessage("Welcome " + student + " the passphrase is " + magic + "\n");
                String input = null;

                // read a message from the client
                while ((input = in.readLine()) != null) {

                    // check to see if the user typed in the "magic" phrase
                    if (input.indexOf(magic) != -1) {
                        // if user typed the "magic" phrase send congratulation message
                        sendMessage("Congratulations " + student + " you sent the passphrase!\n");
                        System.out.println(student + " sent the passphrase!");
                    } // if user did not type in the "magic" phrase
                    // send message to all clients
                    else {
                        ChatServer.this.sendToAllClients(input + "\n");
                    }
                }
            } catch (Exception e) {
            } finally {
                try {
                    // close streams
                    if (in != null) {
                        in.close();
                    }
                    if (out != null) {
                        out.close();                    // close socket
                    }
                    communicationSocket.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }
                ChatServer.this.connectionClosed(this);
            }
        }

        // constructor
        public Connection(Socket s) {
            communicationSocket = s;
        }

        /*******************************************************************
        This method writes the client's message to the socket.
         *******************************************************************/
        public void sendMessage(String message) {
            try {
                out.write(message);
                out.flush();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
}
